'''`bestdori.missions`

BanG Dream! 任务相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from .utils.utils import API
from .utils.network import Api
from ._settings import settings
from .exceptions import (
    MissionNotExistError
)

# 获取总任务信息
def get_all(index: Literal[0, 5]=5) -> dict[str, dict[str, Any]]:
    '''获取总任务信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有任务 ID `all.0.json`
            `5`: 获取所有已有任务的简洁信息 `all.5.json`

    返回:
        dict[str, dict[str, Any]]: 获取到的总任务信息
    '''
    return Api(API['missions']['all'].format(index), settings.proxy).request('get').json()

# 任务类
class Mission:
    '''任务类

    参数:
        id_ (int): 任务 ID
    '''
    # 初始化
    def __init__(self, id_: int) -> None:
        '''任务类

        参数:
            id_ (int): 任务 ID
        '''
        self.id: int = id_
        '''任务 ID'''
        self._info: dict[str, Any] = {}
        '''任务信息'''
        # 检测 ID 是否存在
        all_id = get_all(0)
        if not str(id_) in all_id.keys():
            raise MissionNotExistError(id_)
        return
    
    # 获取任务信息
    def get_info(self) -> dict[str, Any]:
        '''获取任务信息

        返回:
            dict[str, Any]: 任务详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有任务信息存储
            response = Api(
                API['missions']['info'].format(self.id), settings.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取任务评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取任务评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            category_name='MISSION_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取任务标题
    @property
    def title(self) -> str:
        '''获取任务标题

        返回:
            str: 任务标题
        '''
        info = self.get_info()
        # 获取 title 数据
        if (title := info.get('title', None)) is None:
            raise Exception('无法获取任务标题。')
        # 获取第一个非 None 任务标题
        try:
            return next(filter(lambda x: x is not None, title))
        except StopIteration:
            raise Exception('无法获取任务标题。')
    
    # 获取任务所在服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''获取任务所在服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.get_info()
        # 获取 startAt 数据
        if (start_at := info.get('startAt', None)) is None:
            raise Exception('无法获取任务起始时间。')
        # 根据 startAt 数据判断服务器
        if start_at[0] is not None: return 'jp'
        elif start_at[1] is not None: return 'en'
        elif start_at[2] is not None: return 'tw'
        elif start_at[3] is not None: return 'cn'
        elif start_at[4] is not None: return 'kr'
        else:
            raise Exception('无法获取任务所在服务器。')
    


'''`bestdori.costumes`

BanG Dream! 服装相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from ._settings import settings
from .utils.utils import API, ASSETS
from .utils.network import Api, Assets
from .exceptions import (
    CostumeNotExistError,
    AssetsNotExistError
)

# 获取总服装信息
def get_all(index: Literal[0, 5]=5) -> dict[str, dict[str, Any]]:
    '''获取总服装信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有服装 ID `all.0.json`
            `5`: 获取所有已有服装的简洁信息 `all.5.json`

    返回:
        dict[str, dict[str, Any]]: 获取到的总服装信息
    '''
    return Api(API['costumes']['all'].format(index), settings.proxy).request('get').json()

# 服装类
class Costume:
    '''服装类

    参数:
        id_ (int): 服装 ID
    '''
    # 初始化
    def __init__(self, id_: int) -> None:
        '''服装类

        参数:
            id_ (int): 服装 ID
        '''
        self.id: int = id_
        '''服装 ID'''
        self._info: dict[str, Any] = {}
        '''服装信息'''
        # 检测 ID 是否存在
        all_ = get_all(0)
        if not str(id_) in all_.keys():
            raise CostumeNotExistError(id_)
        return
    
    # 获取服装信息
    def get_info(self) -> dict[str, Any]:
        '''获取服装信息

        返回:
            dict[str, Any]: 服装详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有服装信息存储
            response = Api(
                API['costumes']['info'].format(self.id), settings.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取服装评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取服装评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            category_name='COSTUME_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取角色 ID
    @property
    def character_id(self) -> int:
        '''获取角色 ID'''
        info = self.get_info()
        # 获取 characterId 数据
        if (character_id := info.get('characterId', None)) is None:
            raise Exception('无法获取角色 ID。')
        return character_id
    
    # 获取卡牌 ID
    @property
    def card_id(self) -> int:
        '''获取角色 ID'''
        info = self.get_info()
        # 获取 cards 数据
        if (cards := info.get('cards', None)) is None:
            raise Exception('无法获取卡牌 ID。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(self.server)
        return cards[index]
    
    # 获取服装标题
    @property
    def description(self) -> str:
        '''获取服装标题

        返回:
            str: 服装标题
        '''
        info = self.get_info()
        # 获取 description 数据
        if (description := info.get('description', None)) is None:
            raise Exception('无法获取服装标题。')
        # 获取第一个非 None 服装标题
        try:
            return next(filter(lambda x: x is not None, description))
        except StopIteration:
            raise Exception('无法获取服装标题。')
    
    # 获取服装所在服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''获取服装所在服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.get_info()
        # 获取 publishedAt 数据
        if (published_at := info.get('publishedAt', None)) is None:
            raise Exception('无法获取服装发布时间。')
        # 根据 publishedAt 数据判断服务器
        if published_at[0] is not None: return 'jp'
        elif published_at[1] is not None: return 'en'
        elif published_at[2] is not None: return 'tw'
        elif published_at[3] is not None: return 'cn'
        elif published_at[4] is not None: return 'kr'
        else:
            raise Exception('无法获取服装所在服务器。')
    
    # 获取服装模型数据
    @property
    def build_data(self) -> bytes:
        '''获取服装模型数据

        返回:
            bytes: 服装模型数据
        '''
        # 获取服装数据包名称
        info = self.get_info()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取服装数据包名称。')
        try:
            return Assets(
                ASSETS['live2d']['buildData'].format(
                    asset_bundle_name=asset_bundle_name
                ), self.server, settings.proxy
            ).get()
        except AssetsNotExistError:
            raise AssetsNotExistError(f'服装模型 {asset_bundle_name}-{self.server}')
    
    # 获取服装图标
    @property
    def icon(self) -> bytes:
        '''获取服装图标

        返回:
            bytes: 服装图标
        '''
        # 获取服装数据包名称
        info = self.get_info()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取服装数据包名称。')
        try:
            return Assets(
                ASSETS['thumb']['costume'].format(
                    id=str(int(self.id) // 50), asset_bundle_name=asset_bundle_name
                ), self.server, settings.proxy
            ).get()
        except AssetsNotExistError:
            raise AssetsNotExistError(f'服装图标 {asset_bundle_name}-{self.server}')



'''`bestdori.utils.network`

向 Bestdori 发送请求相关模块'''
from json import dumps
from io import BufferedReader
from httpx._models import Cookies
from httpx import Response, Request, Client
from typing import Optional, Literal, cast, Any

from ..exceptions import (
    AssetsNotExistError,
    RequestException,
    REQUEST_EXCEPTION
)

# 向 Bestdori 发送 API 请求类
class Api:
    '''向 Bestdori 发送 API 请求类

    参数:
        api (str): 请求的 API 地址
        
        proxy (Optional[str]): 代理服务器'''
    api: str
    '''请求的 API 地址'''
    proxy: Optional[str]=None
    '''代理服务器'''
    headers: dict[str, str]
    '''请求头'''
    # 初始化
    def __init__(
        self,
        api: str,
        proxy: Optional[str]=None
    ) -> None:
        '''初始化'''
        self.api = api
        self.proxy = proxy
        self.headers = {'Content-Type': 'application/json;charset=UTF-8'}
        return
    
    # 请求发送
    def request(
        self,
        method: Literal['get', 'post'],
        *,
        cookies: Optional[Cookies]=None,
        params: Optional[dict[str, Any]]=None,
        data: Optional[dict[str, Any]]=None,
        files: Optional[dict[str, tuple[str, BufferedReader]]]=None
    ) -> Response:
        '''请求发送

        参数:
            method (Literal[&#39;get&#39;, &#39;post&#39;]): API 调用方法
            
            cookies (Optional[Cookies], optional): Cookies
            
            params (Optional[dict[str, Any]], optional): 调用参数
            
            data (Optional[dict[str, Any]], optional): 调用参数，将以 `json` 字符串形式发送
            
            files (Optional[dict[str, tuple[str, BufferedReader]]], optional): 发送文件参数

        返回:
            Response: 收到的响应
        '''
        # 处理接收到的 API
        if self.api.startswith('http://') or self.api.startswith('https://'):
            self.api = self.api
        else:
            self.api = 'https://bestdori.com/api/' + self.api
        # 构建一个请求体
        request = Request(
            method,
            self.api,
            cookies=cookies,
            params=params,
            data=cast(dict, dumps(data)) if data is not None else data,
            files=files,
            headers=self.headers if not self.api.endswith('/upload') else None
        )
        # 构建代理服务器字典
        if self.proxy is not None:
            proxies = {'http://': self.proxy, 'https://': self.proxy}
        else:
            proxies = None
        
        # 发送请求并获取响应
        with Client(proxies=cast(dict, proxies)) as client:
            response = client.send(request)
            client.close()
        
        # 处理接收到的响应
        response.raise_for_status()
        # 判断接收到的响应是否为 json 格式
        if 'application/json' not in (content_type := response.headers.get('content-type', None)):
            if content_type is not None:
                return response
            else:
                raise Exception('接收到的响应没有 content-type。')
        
        if isinstance((response_data := response.json()), dict):
            if (result := response_data.get('result', None)) is not None:
                if result is False:
                    if (code := response_data.get('code', None)) is not None:
                        if code in REQUEST_EXCEPTION.keys(): # 若错误码已被记录
                            exception_class = REQUEST_EXCEPTION[code]
                            if params is not None:
                                raise exception_class(self.api, **params)
                            elif data is not None:
                                raise exception_class(self.api, **data)
                            else:
                                raise exception_class(self.api)
                        else:
                            raise RequestException(self.api, code)
                    else:
                        raise RequestException(self.api)
        return response

# 获取 Bestdori 资源数据
class Assets:
    '''获取 Bestdori 资源数据

    参数:
        url (str): 请求的资源地址
        
        server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 资源所在服务器
        
        proxy (Optional[str]): 代理服务器'''
    url: str
    '''请求的资源地址'''
    server: Literal['jp', 'en', 'tw', 'cn', 'kr', 'llsif']
    '''资源所在服务器'''
    proxy: Optional[str]=None
    '''代理服务器'''
    # 初始化
    def __init__(
        self,
        url: str,
        server: Literal['jp', 'en', 'tw', 'cn', 'kr', 'llsif'],
        proxy: Optional[str]=None
    ) -> None:
        '''获取 Bestdori 资源数据

        参数:
            url (str): 请求的资源地址
            
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;, &#39;llsif&#39;]): 资源所在服务器
            
            proxy (Optional[str]): 代理服务器
        '''
        self.url = url
        self.server = server
        self.proxy = proxy
        return
    
    # 获取资源连接
    def get_url(self) -> str:
        '''获取资源连接

        返回:
            str: 获取的资源连接 `str`
        '''
        # 如果服务器为 llsif 则转接方法
        if self.server == 'llsif':
            return self._get_niconi_url()
        
        # 处理接收到的 URL
        if self.url.startswith('http://') or self.url.startswith('https://'):
            self.url = self.url
        else:
            self.url = f'https://bestdori.com/assets/{self.server}/' + self.url
        return self.url
    
    # 从 card.niconi.co.ni 获取资源连接
    def _get_niconi_url(self) -> str:
        '''从 card.niconi.co.ni 获取资源连接

        返回:
            str: 获取的资源连接 `str`
        '''
        # 处理接收到的 URL
        if self.url.startswith('http://') or self.url.startswith('https://'):
            self.url = self.url
        else:
            self.url = f'https://card.niconi.co.ni/asset/' + self.url
        return self.url
    
    # 获取资源
    def get(self) -> bytes:
        '''获取资源

        返回:
            bytes: 获取的资源字节数据 `bytes`
        '''
        # 如果服务器为 llsif 则转接方法
        if self.server == 'llsif':
            return self._get_from_niconi()
        
        # 处理接收到的 URL
        if self.url.startswith('http://') or self.url.startswith('https://'):
            self.url = self.url
        else:
            self.url = f'https://bestdori.com/assets/{self.server}/' + self.url
        # 构建一个请求体
        request = Request('get', self.url)
        # 构建代理服务器字典
        if self.proxy is not None:
            proxies = {'http://': self.proxy, 'https://': self.proxy}
        else:
            proxies = None
        
        # 发送请求并获取响应
        with Client(proxies=cast(dict, proxies)) as client:
            response = client.send(request)
            client.close()
        
        response.raise_for_status()
        # 检测响应资源是否存在
        content_type = response.headers.get('content-type', None)
        if content_type is None or content_type == 'text/html':
            raise AssetsNotExistError(self.url)
        return response.content
    
    # 从 card.niconi.co.ni 获取资源
    def _get_from_niconi(self) -> bytes:
        '''从 card.niconi.co.ni 获取资源

        返回:
            bytes: 获取的资源字节数据 `bytes`
        '''
        # 处理接收到的 URL
        if self.url.startswith('http://') or self.url.startswith('https://'):
            self.url = self.url
        else:
            self.url = f'https://card.niconi.co.ni/asset/' + self.url
        # 构建一个请求体
        request = Request('get', self.url)
        # 构建代理服务器字典
        if self.proxy is not None:
            proxies = {'http://': self.proxy, 'https://': self.proxy}
        else:
            proxies = None
        
        # 发送请求并获取响应
        with Client(proxies=cast(dict, proxies)) as client:
            response = client.send(request)
            client.close()
        
        response.raise_for_status()
        # 检测响应资源是否存在
        content_type = response.headers.get('content-type', None)
        if content_type is None or content_type == 'text/html':
            raise AssetsNotExistError(self.url)
        return response.content

# 获取 Bestdori res 资源数据
class Res:
    '''获取 Bestdori res 资源数据

    参数:
        url (str): 请求的 res 资源地址
        
        proxy (Optional[str]): 代理服务器'''
    url: str
    '''请求的资源地址'''
    proxy: Optional[str]=None
    '''代理服务器'''
    # 初始化
    def __init__(
        self,
        url: str,
        proxy: Optional[str]=None
    ) -> None:
        '''获取 Bestdori 资源数据

        参数:
            url (str): 请求的资源地址
            
            proxy (Optional[str]): 代理服务器
        '''
        self.url = url
        self.proxy = proxy
        return
    
    # 获取资源
    def get(self) -> bytes:
        '''获取资源

        返回:
            bytes: 获取的资源字节数据 `bytes`
        '''
        # 处理接收到的 URL
        if self.url.startswith('http://') or self.url.startswith('https://'):
            self.url = self.url
        else:
            self.url = f'https://bestdori.com/res/' + self.url
        # 构建一个请求体
        request = Request(
            'get',
            self.url
        )
        # 构建代理服务器字典
        if self.proxy is not None:
            proxies = {'http://': self.proxy, 'https://': self.proxy}
        else:
            proxies = None
        
        # 发送请求并获取响应
        with Client(proxies=cast(dict, proxies)) as client:
            response = client.send(request)
            client.close()
        
        response.raise_for_status()
        # 检测响应资源是否存在
        content_type = response.headers.get('content-type', None)
        if content_type is None or content_type == 'text/html':
            raise AssetsNotExistError(self.url)
        return response.content
